
FruitPreparerMultiFruit = {}
FruitPreparerMultiFruit.modDir = g_currentModDirectory
FruitPreparerMultiFruit.modName = g_currentModName
FruitPreparerMultiFruit.usedXmlString = "mapFruitTypes_fruitPrearer"
FruitPreparerMultiFruit.fruitTypesString = ""



function FruitPreparerMultiFruit.initSpecialization()

	g_workAreaTypeManager:addWorkAreaType("fruitPreparer", false)

end



function FruitPreparerMultiFruit.prerequisitesPresent(specializations)

	return SpecializationUtil.hasSpecialization(WorkArea, specializations) and SpecializationUtil.hasSpecialization(TurnOnVehicle, specializations)

end



function FruitPreparerMultiFruit.registerFunctions(vehicleType)

	SpecializationUtil.registerFunction(vehicleType, "onFieldDataUpdateFinished", FruitPreparerMultiFruit.onFieldDataUpdateFinished)
	SpecializationUtil.registerFunction(vehicleType, "processFruitPreparerArea", FruitPreparerMultiFruit.processFruitPreparerArea)

end



function FruitPreparerMultiFruit.registerOverwrittenFunctions(vehicleType)

	SpecializationUtil.registerOverwrittenFunction(vehicleType, "loadWorkAreaFromXML",					FruitPreparerMultiFruit.loadWorkAreaFromXML)
	SpecializationUtil.registerOverwrittenFunction(vehicleType, "getDoGroundManipulation",				FruitPreparerMultiFruit.getDoGroundManipulation)
	SpecializationUtil.registerOverwrittenFunction(vehicleType, "doCheckSpeedLimit",					FruitPreparerMultiFruit.doCheckSpeedLimit)
	SpecializationUtil.registerOverwrittenFunction(vehicleType, "getAllowCutterAIFruitRequirements",	FruitPreparerMultiFruit.getAllowCutterAIFruitRequirements)
	SpecializationUtil.registerOverwrittenFunction(vehicleType, "getDirtMultiplier",					FruitPreparerMultiFruit.getDirtMultiplier)
	SpecializationUtil.registerOverwrittenFunction(vehicleType, "getWearMultiplier",					FruitPreparerMultiFruit.getWearMultiplier)

end



function FruitPreparerMultiFruit.registerEventListeners(vehicleType)

	SpecializationUtil.registerEventListener(vehicleType, "onLoad", FruitPreparerMultiFruit)
	SpecializationUtil.registerEventListener(vehicleType, "onPostLoad", FruitPreparerMultiFruit)
	SpecializationUtil.registerEventListener(vehicleType, "onDelete", FruitPreparerMultiFruit)
	SpecializationUtil.registerEventListener(vehicleType, "onReadUpdateStream", FruitPreparerMultiFruit)
	SpecializationUtil.registerEventListener(vehicleType, "onWriteUpdateStream", FruitPreparerMultiFruit)
	SpecializationUtil.registerEventListener(vehicleType, "onTurnedOn", FruitPreparerMultiFruit)
	SpecializationUtil.registerEventListener(vehicleType, "onTurnedOff", FruitPreparerMultiFruit)
	SpecializationUtil.registerEventListener(vehicleType, "onStartWorkAreaProcessing", FruitPreparerMultiFruit)

end



function FruitPreparerMultiFruit:onLoad(savegame)

	local spec = self["spec_" .. FruitPreparerMultiFruit.modName .. ".FruitPreparerMultiFruit"]

	XMLUtil.checkDeprecatedXMLElements(self.xmlFile, self.configFileName, "vehicle.turnOnAnimation#name", "vehicle.turnOnVehicle.turnedAnimation#name") -- FS15 to FS17
	XMLUtil.checkDeprecatedXMLElements(self.xmlFile, self.configFileName, "vehicle.turnOnAnimation#speed", "vehicle.turnOnVehicle.turnedAnimation#turnOnSpeedScale") -- FS15 to FS17
	XMLUtil.checkDeprecatedXMLElements(self.xmlFile, self.configFileName, "vehicle.fruitPreparer#useReelStateToTurnOn") -- FS17 to FS19
	XMLUtil.checkDeprecatedXMLElements(self.xmlFile, self.configFileName, "vehicle.fruitPreparer#onlyActiveWhenLowered") -- FS17 to FS19
	XMLUtil.checkDeprecatedXMLElements(self.xmlFile, self.configFileName, "vehicle.vehicle.fruitPreparerSound", "vehicle.fruitPreparer.sounds.work") -- FS17 to FS19
	XMLUtil.checkDeprecatedXMLElements(self.xmlFile, self.configFileName, "vehicle.turnedOnRotationNodes.turnedOnRotationNode", "vehicle.fruitPreparer.animationNodes.animationNode", "fruitPreparer") -- FS17 to FS19

	if self.isClient then
		spec.samples = {}
		spec.samples.work = g_soundManager:loadSampleFromXML(self.xmlFile, "vehicle.fruitPreparer.sounds", "work", self.baseDirectory, self.components, 0, AudioGroup.VEHICLE, self.i3dMappings, self)
		spec.animationNodes = g_animationManager:loadAnimations(self.xmlFile, "vehicle.fruitPreparer.animationNodes", self.components, self, self.i3dMappings)
	end

	spec.fruitType = FruitType.UNKNOWN

	spec.fruitTypes = {}
	local fruitTypesString = getXMLString(self.xmlFile, "vehicle.fruitPreparer#fruitType")
	local i = 0

	if fruitTypesString == FruitPreparerMultiFruit.usedXmlString then

		fruitTypesString = ""

		for fruitTypeIndex, fruitTypeData in pairs(g_fruitTypeManager.fruitTypes) do

			if fruitTypeData ~= nil then

				local minHarvestingGrowthState = fruitTypeData.minHarvestingGrowthState
				local minPreparingGrowthState = fruitTypeData.minPreparingGrowthState
				local preparedGrowthState = fruitTypeData.preparedGrowthState
				local name = fruitTypeData.name

				if minHarvestingGrowthState == 9 and minPreparingGrowthState ~= -1 and preparedGrowthState == 9 and name ~= "SUGARCANE" then

					if fruitTypesString ~= "" then 
						fruitTypesString = fruitTypesString .." "
					end

					fruitTypesString = fruitTypesString..tostring(name)

				end

			end

		end

	end

	FruitPreparerMultiFruit.fruitTypesString = fruitTypesString

	for substring in fruitTypesString:gmatch("%S+") do

		local fruitIndex = g_fruitTypeManager:getFruitTypeByName(substring)

		if fruitIndex ~= nil then
			spec.fruitTypes[i] = fruitIndex.index
			i = i + 1
		end

	end




	if #spec.fruitTypes > 0 then

		for _, index in pairs(spec.fruitTypes) do

			local desc = g_fruitTypeManager:getFruitTypeByIndex(index)

			if desc == nil then
				g_logManager:xmlWarning(self.configFileName, "Unable to find fruitTypeIndex '%d' in fruitPreparer", index)
			end

		end

	else
		g_logManager:xmlWarning(self.configFileName, "Missing fruitTypes in fruitPreparer")
	end

	spec.isWorking = false

end



function FruitPreparerMultiFruit:onPostLoad()

	for _, item in pairs(g_storeManager.items) do

		if item.customEnvironment == FruitPreparerMultiFruit.modName then

			if item.specs.fillTypes.fillTypeNames == FruitPreparerMultiFruit.usedXmlString then

				item.specs.fillTypes.fillTypeNames = FruitPreparerMultiFruit.fruitTypesString

			end

			if item.specs.fillTypes.fruitTypeNames == FruitPreparerMultiFruit.usedXmlString then

				item.specs.fillTypes.fruitTypeNames = FruitPreparerMultiFruit.fruitTypesString

			end

		end

	end

end



function FruitPreparerMultiFruit:onDelete()

	if self.isClient then

		local spec = self["spec_" .. FruitPreparerMultiFruit.modName .. ".FruitPreparerMultiFruit"]

		g_soundManager:deleteSamples(spec.samples)
		g_animationManager:deleteAnimations(spec.animationNodes)

	end

end



function FruitPreparerMultiFruit:onReadUpdateStream(streamId, timestamp, connection)

	if connection:getIsServer() then

		local spec = self["spec_" .. FruitPreparerMultiFruit.modName .. ".FruitPreparerMultiFruit"]

		spec.isWorking = streamReadBool(streamId)

	end

end



function FruitPreparerMultiFruit:onWriteUpdateStream(streamId, connection, dirtyMask)

	if not connection:getIsServer() then

		local spec = self["spec_" .. FruitPreparerMultiFruit.modName .. ".FruitPreparerMultiFruit"]

		streamWriteBool(streamId, spec.isWorking)

	end

end



function FruitPreparerMultiFruit:onTurnedOn()

	if self.isClient then

		local spec = self["spec_" .. FruitPreparerMultiFruit.modName .. ".FruitPreparerMultiFruit"]

		g_soundManager:playSample(spec.samples.work)
		g_animationManager:startAnimations(spec.animationNodes)

	end

end



function FruitPreparerMultiFruit:onTurnedOff()

	if self.isClient then

		local spec = self["spec_" .. FruitPreparerMultiFruit.modName .. ".FruitPreparerMultiFruit"]

		g_soundManager:stopSamples(spec.samples)
		g_animationManager:stopAnimations(spec.animationNodes)

	end

end



function FruitPreparerMultiFruit:loadWorkAreaFromXML(superFunc, workArea, xmlFile, key)

	local retValue = superFunc(self, workArea, xmlFile, key)

	if workArea.type == WorkAreaType.FRUITPREPARER then

		XMLUtil.checkDeprecatedXMLElements(self.xmlFile, self.configFileName, key.."#dropStartIndex", key..".fruitPreparer#dropWorkAreaIndex") -- FS17 to FS19
		XMLUtil.checkDeprecatedXMLElements(self.xmlFile, self.configFileName, key.."#dropWidthIndex", key..".fruitPreparer#dropWorkAreaIndex") -- FS17 to FS19
		XMLUtil.checkDeprecatedXMLElements(self.xmlFile, self.configFileName, key.."#dropHeightIndex", key..".fruitPreparer#dropWorkAreaIndex") -- FS17 to FS19
		workArea.dropWorkAreaIndex = getXMLInt(xmlFile, key..".fruitPreparer#dropWorkAreaIndex")

	end

	return retValue

end



function FruitPreparerMultiFruit:getDoGroundManipulation(superFunc)

	local spec = self["spec_" .. FruitPreparerMultiFruit.modName .. ".FruitPreparerMultiFruit"]

	return superFunc(self) and spec.isWorking

end



function FruitPreparerMultiFruit:doCheckSpeedLimit(superFunc)

	return superFunc(self) or (self:getIsTurnedOn() and (self.getIsImplementChainLowered == nil or self:getIsImplementChainLowered()))

end



function FruitPreparerMultiFruit:getAllowCutterAIFruitRequirements(superFunc)

	return false

end



function FruitPreparerMultiFruit:getDirtMultiplier(superFunc)

	local spec = self["spec_" .. FruitPreparerMultiFruit.modName .. ".FruitPreparerMultiFruit"]

	if spec.isWorking then

		return superFunc(self) + self:getWorkDirtMultiplier() * self:getLastSpeed() / self.speedLimit

	end

	return superFunc(self)

end



function FruitPreparerMultiFruit:getWearMultiplier(superFunc)

	local spec = self["spec_" .. FruitPreparerMultiFruit.modName .. ".FruitPreparerMultiFruit"]

	if spec.isWorking then

		return superFunc(self) + self:getWorkWearMultiplier() * self:getLastSpeed() / self.speedLimit

	end

	return superFunc(self)

end



function FruitPreparerMultiFruit.getDefaultSpeedLimit()

	return 15

end



function FruitPreparerMultiFruit:onStartWorkAreaProcessing(dt)

	local spec = self["spec_" .. FruitPreparerMultiFruit.modName .. ".FruitPreparerMultiFruit"]

	spec.isWorking = false

end



function FruitPreparerMultiFruit:processFruitPreparerArea(workArea)

	local spec = self["spec_" .. FruitPreparerMultiFruit.modName .. ".FruitPreparerMultiFruit"]
	local workAreaSpec = self.spec_workArea
	local xs,_,zs = getWorldTranslation(workArea.start)
	local xw,_,zw = getWorldTranslation(workArea.width)
	local xh,_,zh = getWorldTranslation(workArea.height)
	local dxs, dzs = xs, zs
	local dxw, dzw = xw, zw
	local dxh, dzh = xh, zh

	if workArea.dropWorkAreaIndex ~= nil then

		local dropArea = workAreaSpec.workAreas[workArea.dropWorkAreaIndex]

		if dropArea ~= nil then

			dxs, _, dzs = getWorldTranslation(dropArea.start)
			dxw, _, dzw = getWorldTranslation(dropArea.width)
			dxh, _, dzh = getWorldTranslation(dropArea.height)

		end

	end

	if not self.requestedFieldData then

		self.requestedFieldData = true

		FSDensityMapUtil.getFieldStatusAsync(xs, zs, xw, zw, xh, zh, self.onFieldDataUpdateFinished, self)

	end

	local area = FSDensityMapUtil.updateFruitPreparerArea(spec.fruitType, xs,zs, xw,zw, xh,zh, dxs,dzs, dxw,dzw, dxh,dzh)

	if area > 0 then
		spec.isWorking = true
	end

	return 0, area

end



function FruitPreparerMultiFruit:onFieldDataUpdateFinished(data)

	if not self.requestedFieldData then
		return
	end

	local spec = self["spec_" .. FruitPreparerMultiFruit.modName .. ".FruitPreparerMultiFruit"]

	local hasData = data ~= nil

	if hasData then

		for fruitDescIndex, _ in pairs(data.fruits) do

			if data.fruitPixels[fruitDescIndex] > 0 then

				for _, allowed in pairs(spec.fruitTypes) do

					if fruitDescIndex == allowed then

						spec.fruitType = fruitDescIndex

					end

				end

			end

		end

	end

	self.requestedFieldData = false

end


-- fillUnit overwritten Function

function FruitPreparerMultiFruit.overwrite_fillUnit_loadFillUnitFromXML(self, superFunc, xmlFile, key, entry, index)

	local spec = self.spec_fillUnit

	entry.fillUnitIndex = index
	entry.capacity = Utils.getNoNil(getXMLFloat(xmlFile, key .. "#capacity"), math.huge)
	entry.updateMass = Utils.getNoNil(getXMLBool(xmlFile, key .. "#updateMass"), true)
	entry.canBeUnloaded = Utils.getNoNil(getXMLBool(xmlFile, key .. "#canBeUnloaded"), true)
	entry.needsSaving = true
	entry.fillLevel = 0
	entry.fillLevelSent = 0
	entry.fillType = FillType.UNKNOWN
	entry.fillTypeSent = FillType.UNKNOWN
	entry.fillTypeToDisplay = FillType.UNKNOWN
	entry.fillLevelToDisplay = nil
	entry.lastValidFillType = FillType.UNKNOWN
	entry.lastValidFillTypeSent = FillType.UNKNOWN



	if hasXMLProperty(xmlFile, key .. ".exactFillRootNode") then

		XMLUtil.checkDeprecatedXMLElements(xmlFile, self.configFileName, key .. ".exactFillRootNode#index", key .. ".exactFillRootNode#node") --FS17 to FS19
		entry.exactFillRootNode = I3DUtil.indexToObject(self.components, getXMLString(xmlFile, key .. ".exactFillRootNode#node"), self.i3dMappings)

		if entry.exactFillRootNode ~= nil then

			local colMask = getCollisionMask(entry.exactFillRootNode)

			if bitAND(FillUnit.EXACTFILLROOTNODE_MASK, colMask) == 0 then

				g_logManager:xmlWarning(self.configFileName, "Invalid collision mask for exactFillRootNode '%s'. Bit 30 needs to be set!", key)
				return false

			end

			spec.exactFillRootNodeToFillUnit[entry.exactFillRootNode] = entry
			spec.exactFillRootNodeToExtraDistance[entry.exactFillRootNode] = getXMLFloat(xmlFile, key .. ".exactFillRootNode#extraEffectDistance") or 0
			spec.hasExactFillRootNodes = true

		else

			g_logManager:xmlWarning(self.configFileName, "ExactFillRootNode not found for fillUnit '%s'!", key)

		end

	end



	XMLUtil.checkDeprecatedXMLElements(xmlFile, self.configFileName, key .. ".autoAimTargetNode#index", key .. ".autoAimTargetNode#node") --FS17 to FS19
	entry.autoAimTarget = {}
	entry.autoAimTarget.node = I3DUtil.indexToObject(self.components, getXMLString(xmlFile, key .. ".autoAimTargetNode#node"), self.i3dMappings)

	if entry.autoAimTarget.node ~= nil then

		entry.autoAimTarget.baseTrans = {getTranslation(entry.autoAimTarget.node)}
		entry.autoAimTarget.startZ = getXMLFloat(xmlFile,  key ..".autoAimTargetNode#startZ")
		entry.autoAimTarget.endZ = getXMLFloat(xmlFile,  key ..".autoAimTargetNode#endZ")
		entry.autoAimTarget.startPercentage = Utils.getNoNil(getXMLFloat(xmlFile,  key ..".autoAimTargetNode#startPercentage"), 25)/100
		entry.autoAimTarget.invert = Utils.getNoNil(getXMLBool(xmlFile,  key ..".autoAimTargetNode#invert"), false)

		if entry.autoAimTarget.startZ ~= nil and entry.autoAimTarget.endZ ~= nil then

			local startZ = entry.autoAimTarget.startZ

			if entry.autoAimTarget.invert then

				startZ = entry.autoAimTarget.endZ

			end

			setTranslation(entry.autoAimTarget.node, entry.autoAimTarget.baseTrans[1], entry.autoAimTarget.baseTrans[2], startZ)

		end

	end



	entry.supportedFillTypes = {}

	local fillTypes
	local fillTypeCategories = getXMLString(xmlFile, key .. "#fillTypeCategories")
	local fillTypeNames = getXMLString(xmlFile, key .. "#fillTypes")

	if fillTypeNames == FruitPreparerMultiFruit.usedXmlString then

		fillTypeNames = ""

		for fruitTypeIndex, fruitTypeData in pairs(g_fruitTypeManager.fruitTypes) do

			if fruitTypeData ~= nil then

				local minHarvestingGrowthState = fruitTypeData.minHarvestingGrowthState
				local minPreparingGrowthState = fruitTypeData.minPreparingGrowthState
				local preparedGrowthState = fruitTypeData.preparedGrowthState
				local name = fruitTypeData.name

				if minHarvestingGrowthState == 9 and minPreparingGrowthState ~= -1 and preparedGrowthState == 9 and name ~= "SUGARCANE" then

					if fillTypeNames ~= "" then 
						fillTypeNames = fillTypeNames .." "
					end

					fillTypeNames = fillTypeNames..tostring(name)

				end

			end

		end

	end

	if fillTypeCategories ~= nil and fillTypeNames == nil then

		fillTypes = g_fillTypeManager:getFillTypesByCategoryNames(fillTypeCategories, "Warning: '"..self.configFileName.. "' has invalid fillTypeCategory '%s'.")

	elseif fillTypeCategories == nil and fillTypeNames ~= nil then

		fillTypes = g_fillTypeManager:getFillTypesByNames(fillTypeNames, "Warning: '"..self.configFileName.. "' has invalid fillType '%s'.")

	else

		g_logManager:xmlWarning(self.configFileName, "Missing 'fillTypeCategories' or 'fillTypes' for fillUnit '%s'", key)
		return false

	end

	if fillTypes ~= nil then

		for _,fillType in pairs(fillTypes) do

			entry.supportedFillTypes[fillType] = true

		end

	end



	entry.supportedToolTypes = {}

	for i=1, g_toolTypeManager:getNumberOfToolTypes() do

		entry.supportedToolTypes[i] = true

	end

	local startFillLevel = getXMLFloat(xmlFile, key .. "#startFillLevel")
	local startFillTypeStr = getXMLString(xmlFile, key .. "#startFillType")

	if startFillTypeStr ~= nil then

		local startFillTypeIndex = g_fillTypeManager:getFillTypeIndexByName(startFillTypeStr)

		if startFillTypeIndex ~= nil then

			entry.startFillLevel = startFillLevel
			entry.startFillTypeIndex = startFillTypeIndex

		end

	end

	entry.fillRootNode = I3DUtil.indexToObject(self.components, getXMLString(xmlFile, key .. ".fillRootNode#node"), self.i3dMappings)

	if entry.fillRootNode == nil then

		entry.fillRootNode = self.components[1].node

	end

	entry.fillMassNode = I3DUtil.indexToObject(self.components, getXMLString(xmlFile, key .. ".fillMassNode#node"), self.i3dMappings)

	local updateFillLevelMass = Utils.getNoNil(getXMLBool(xmlFile, key .. "#updateFillLevelMass"), true)

	if entry.fillMassNode == nil and updateFillLevelMass then

		entry.fillMassNode = self.components[1].node

	end



	-- mp sync info
	entry.synchronizeFillLevel = Utils.getNoNil(getXMLBool(xmlFile, key .. "#synchronizeFillLevel"), true)
	entry.synchronizeFullFillLevel = Utils.getNoNil(getXMLBool(xmlFile, key .. "#synchronizeFullFillLevel"), false)

	local defaultBits = 16

	for startCapacity, bits in pairs(FillUnit.CAPACITY_TO_NETWORK_BITS) do

		if entry.capacity >= startCapacity then

			defaultBits = bits

		end

	end

	entry.synchronizationNumBits = Utils.getNoNil(getXMLInt(xmlFile, key .. "#synchronizationNumBits"), defaultBits)
	entry.showOnHud = Utils.getNoNil(getXMLBool(xmlFile, key .. "#showOnHud"), true)
	entry.blocksAutomatedTrainTravel = Utils.getNoNil(getXMLBool(xmlFile, key .. "#blocksAutomatedTrainTravel"), false)
	entry.fillAnimation = getXMLString(xmlFile, key .. "#fillAnimation")
	entry.fillAnimationLoadTime = getXMLFloat(xmlFile, key .. "#fillAnimationLoadTime")
	entry.fillAnimationEmptyTime = getXMLFloat(xmlFile, key .. "#fillAnimationEmptyTime")



	if self.isClient then

		entry.alarmTriggers = {}

		local i = 0

		while true do

			local nodeKey = key .. string.format(".alarmTriggers.alarmTrigger(%d)", i)

			if not hasXMLProperty(xmlFile, nodeKey) then
				break
			end

			local alarmTrigger = {}

			if self:loadAlarmTrigger(xmlFile, nodeKey, alarmTrigger, entry) then

				table.insert(entry.alarmTriggers, alarmTrigger)

			end

			i = i + 1

		end

		entry.measurementNodes = {}

		i = 0

		while true do

			local nodeKey = key .. string.format(".measurementNodes.measurementNode(%d)", i)

			if not hasXMLProperty(xmlFile, nodeKey) then
				break
			end

			local measurementNode = {}

			if self:loadMeasurementNode(xmlFile, nodeKey, measurementNode) then

				table.insert(entry.measurementNodes, measurementNode)

			end

			i = i + 1

		end

		entry.fillPlane = {}
		entry.lastFillPlaneType = nil

		if not self:loadFillPlane(xmlFile, key..".fillPlane", entry.fillPlane, entry) then

			entry.fillPlane = nil

		end

		entry.fillEffects = g_effectManager:loadEffect(xmlFile, key .. ".fillEffect", self.components, self, self.i3dMappings)
		entry.animationNodes = g_animationManager:loadAnimations(xmlFile, key..".animationNodes", self.components, self, self.i3dMappings)
		XMLUtil.checkDeprecatedXMLElements(xmlFile, self.configFileName, key .. ".fillLevelHud", key .. ".dashboard") --FS17 to FS19

		if self.loadDashboardsFromXML ~= nil then

			self:loadDashboardsFromXML(xmlFile, key, {valueTypeToLoad = "fillLevel",
													  valueObject = entry,
													  valueFunc = "fillLevel",
													  minFunc = 0,
													  maxFunc = "capacity"})

			self:loadDashboardsFromXML(xmlFile, key, {valueTypeToLoad = "fillLevelPct",
													  valueObject = entry,
													  valueFunc = function(fillUnit)
														  return MathUtil.clamp(fillUnit.fillLevel/fillUnit.capacity, 0, 1) * 100
													  end,
													  minFunc = 0,
													  maxFunc = 100})

			self:loadDashboardsFromXML(xmlFile, key, {valueTypeToLoad = "fillLevelWarning",
													  valueObject = entry,
													  valueFunc = "fillLevel",
													  minFunc = 0,
													  maxFunc = "capacity",
													  additionalAttributesFunc = Dashboard.warningAttributes,
													  stateFunc = Dashboard.warningState})

		end

	end

	return true

end

FillUnit.loadFillUnitFromXML = Utils.overwrittenFunction(FillUnit.loadFillUnitFromXML, FruitPreparerMultiFruit.overwrite_fillUnit_loadFillUnitFromXML )



-- cutter overwritten Function

function FruitPreparerMultiFruit.overwrite_cutter_onLoad(self, superFunc, savegame)

	local spec = self.spec_cutter

	XMLUtil.checkDeprecatedXMLElements(self.xmlFile, self.configFileName, "vehicle.turnedOnRotationNodes.turnedOnRotationNode#type", "vehicle.cutter.animationNodes.animationNode", "cutter") --FS17 to FS19
	XMLUtil.checkDeprecatedXMLElements(self.xmlFile, self.configFileName, "vehicle.turnedOnScrollers", "vehicle.cutter.animationNodes.animationNode") --FS17 to FS19
	XMLUtil.checkDeprecatedXMLElements(self.xmlFile, self.configFileName, "vehicle.cutter.turnedOnScrollers", "vehicle.cutter.animationNodes.animationNode") --FS17 to FS19
	XMLUtil.checkDeprecatedXMLElements(self.xmlFile, self.configFileName, "vehicle.cutter.reelspikes", "vehicle.cutter.rotationNodes.rotationNode or vehicle.turnOnVehicle.turnedOnAnimation") --FS17 to FS19
	XMLUtil.checkDeprecatedXMLElements(self.xmlFile, self.configFileName, "vehicle.cutter.threshingParticleSystems.threshingParticleSystem", "vehicle.cutter.fillEffect.effectNode") --FS17 to FS19
	XMLUtil.checkDeprecatedXMLElements(self.xmlFile, self.configFileName, "vehicle.cutter.threshingParticleSystems.emitterShape", "vehicle.cutter.fillEffect.effectNode") --FS17 to FS19
	XMLUtil.checkDeprecatedXMLElements(self.xmlFile, self.configFileName, "vehicle.cutter#convertedFillTypeCategories", "vehicle.cutter#fruitTypeConverter") --FS17 to FS19
	XMLUtil.checkDeprecatedXMLElements(self.xmlFile, self.configFileName, "vehicle.cutter#startAnimationName", "vehicle.turnOnVehicle.turnOnAnimation#name") --FS17 to FS19



	-- load fruitTypes
	local fruitTypes = nil
	local fruitTypeNames = getXMLString(self.xmlFile, "vehicle.cutter#fruitTypes")
	local fruitTypeCategories = getXMLString(self.xmlFile, "vehicle.cutter#fruitTypeCategories")

	if fruitTypeNames == FruitPreparerMultiFruit.usedXmlString then

		fruitTypeNames = ""

		for fruitTypeIndex, fruitTypeData in pairs(g_fruitTypeManager.fruitTypes) do

			if fruitTypeData ~= nil then

				local minHarvestingGrowthState = fruitTypeData.minHarvestingGrowthState
				local minPreparingGrowthState = fruitTypeData.minPreparingGrowthState
				local preparedGrowthState = fruitTypeData.preparedGrowthState
				local name = fruitTypeData.name

				if minHarvestingGrowthState == 9 and minPreparingGrowthState ~= -1 and preparedGrowthState == 9 and name ~= "SUGARCANE" then

					if fruitTypeNames ~= "" then 
						fruitTypeNames = fruitTypeNames .." "
					end

					fruitTypeNames = fruitTypeNames..tostring(name)

				end

			end

		end

	end

	if fruitTypeCategories ~= nil and fruitTypeNames == nil then

		fruitTypes = g_fruitTypeManager:getFruitTypesByCategoryNames(fruitTypeCategories, "Warning: Cutter has invalid fruitTypeCategory '%s' in '"..self.configFileName.."'")

	elseif fruitTypeCategories == nil and fruitTypeNames ~= nil then

		fruitTypes = g_fruitTypeManager:getFruitTypesByNames(fruitTypeNames, "Warning: Cutter has invalid fruitType '%s' in '"..self.configFileName.."'")

	else

		g_logManager:xmlWarning(self.configFileName, "Cutter needs either the 'fruitTypeCategories' or 'fruitTypes' attribute!")

	end

	if fruitTypes ~= nil then

		spec.fruitTypes = {}

		for _,fruitType in pairs(fruitTypes) do

			spec.fruitTypes[fruitType] = true

		end

	end



	spec.fruitTypeConverters = {}

	local category = getXMLString(self.xmlFile, "vehicle.cutter#fruitTypeConverter")

	if category ~= nil then

		local data = g_fruitTypeManager:getConverterDataByName(category)

		if data ~= nil then

			for input, converter in pairs(data) do

				spec.fruitTypeConverters[input] = converter

			end

		end

	end

	if self.isClient then

		spec.animationNodes = g_animationManager:loadAnimations(self.xmlFile, "vehicle.cutter.animationNodes", self.components, self, self.i3dMappings)
		spec.fruitExtraObjects = {}

		local i = 0

		while true do

			local key = string.format("vehicle.cutter.fruitExtraObjects.fruitExtraObject(%d)", i)

			XMLUtil.checkDeprecatedXMLElements(self.xmlFile, self.configFileName, key.."#index", key.."#node")

			local node = I3DUtil.indexToObject(self.components, getXMLString(self.xmlFile, key.."#node"), self.i3dMappings)
			local anim = getXMLString(self.xmlFile, key.."#anim")
			local isDefault = Utils.getNoNil(getXMLBool(self.xmlFile, key.."#isDefault"), false)
			local fruitType = g_fruitTypeManager:getFruitTypeByName(getXMLString(self.xmlFile, key.."#fruitType"))

			if fruitType == nil or (node == nil and anim == nil) then
				break
			end

			if node ~= nil then
				setVisibility(node, false)
			end

			local extraObject = {node=node, anim=anim}

			spec.fruitExtraObjects[fruitType.index] = extraObject

			if isDefault then
				spec.fruitExtraObjects[FruitType.UNKNOWN] = extraObject
			end

			i = i + 1

		end

		spec.hideExtraObjectsOnDetach = Utils.getNoNil(getXMLBool(self.xmlFile, "vehicle.cutter.fruitExtraObjects#hideOnDetach"), false)
		spec.testAreas = {}
		spec.testAreaBase = Utils.getNoNil(I3DUtil.indexToObject(self.components, getXMLString(self.xmlFile, "vehicle.cutter.testAreas#baseNode"), self.i3dMappings), self.rootNode)

		i = 0

		while true do

			local key = string.format("vehicle.cutter.testAreas.testArea(%d)", i)

			if not hasXMLProperty(self.xmlFile, key) then
				break
			end

			local testArea = {}

			if self:loadTestAreaFromXML(testArea, self.xmlFile, key) then
				table.insert(spec.testAreas, testArea)
			end

			i = i + 1

		end



		-- cutter
		spec.cutterEffects = {}
		spec.currentCutterEffect = nil

		if hasXMLProperty(self.xmlFile, "vehicle.cutter.effect") then

			for fruitTypeIndex,_ in pairs(spec.fruitTypes) do

				spec.cutterEffects[fruitTypeIndex] = {}

				i = 0

				while true do

					local key = string.format("vehicle.cutter.effect.effectNode(%d)", i)

					if not hasXMLProperty(self.xmlFile, key) then
						break
					end

					local effect = CutterEffect:new()

					effect:load(self.xmlFile, key, self.components, self, fruitTypeIndex, self.i3dMappings)

					table.insert(spec.cutterEffects[fruitTypeIndex], effect)

					i = i + 1

				end

			end

		end

		spec.fillEffects = g_effectManager:loadEffect(self.xmlFile, "vehicle.cutter.fillEffect", self.components, self, self.i3dMappings)

	end



	spec.allowsForageGrowthState = Utils.getNoNil(getXMLBool(self.xmlFile, "vehicle.cutter#allowsForageGrowthState"), false)
	spec.allowCuttingWhileRaised = Utils.getNoNil(getXMLBool(self.xmlFile, "vehicle.cutter#allowCuttingWhileRaised"), false)
	spec.movingDirection = MathUtil.sign(Utils.getNoNil(getXMLInt(self.xmlFile, "vehicle.cutter#movingDirection"), 1))
	spec.strawRatio = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.cutter#strawRatio"), 1)
	spec.useWindrow = false
	spec.currentInputFillType = FillType.UNKNOWN
	spec.currentInputFruitType = FruitType.UNKNOWN
	spec.currentInputFruitTypeAI = FruitType.UNKNOWN
	spec.lastValidInputFruitType = FruitType.UNKNOWN
	spec.currentInputFruitTypeSent = FruitType.UNKNOWN
	spec.currentGrowthStateArea = 0
	spec.currentGrowthState = 0
	spec.lastAreaBiggerZero = false
	spec.lastAreaBiggerZeroSent = false
	spec.lastAreaBiggerZeroTime = -1
	spec.workAreaParameters = {}
	spec.workAreaParameters.lastRealArea = 0
	spec.workAreaParameters.lastArea = 0
	spec.workAreaParameters.lastGrowthState = 0
	spec.workAreaParameters.lastGrowthStateArea = 0
	spec.lastOutputFillTypes = {}
	spec.lastPrioritizedOutputType = FillType.UNKNOWN
	spec.lastOutputTime = 0
	spec.cutterLoad = 0
	spec.isWorking = false
	spec.workAreaParameters.countArea = true
	spec.aiNoValidGroundTimer = 0
	spec.dirtyFlag = self:getNextDirtyFlag()
	spec.effectDirtyFlag = self:getNextDirtyFlag()

	self:doCollisionMaskCheck(MathUtil.bitsToMask(26), "vehicle.cutter#ignoreCollisionMask")

end

Cutter.onLoad = Utils.overwrittenFunction(Cutter.onLoad, FruitPreparerMultiFruit.overwrite_cutter_onLoad )



function FruitPreparerMultiFruit.overwrite_cutter_processCutterArea(self, superFunc, workArea, dt)

	local spec = self.spec_cutter

	if spec.workAreaParameters.combineVehicle ~= nil then

		local xs,_,zs = getWorldTranslation(workArea.start)
		local xw,_,zw = getWorldTranslation(workArea.width)
		local xh,_,zh = getWorldTranslation(workArea.height)
		local lastRealArea = 0
		local lastThreshedArea = 0
		local lastArea = 0

		for fruitType, state in pairs(spec.workAreaParameters.fruitTypesToUse) do

			if state then

				local realArea, area, sprayFactor, plowFactor, limeFactor, weedFactor, growthState, growthStateArea, terrainDetailPixelsSum = FSDensityMapUtil.cutFruitArea(fruitType, xs,zs, xw,zw, xh,zh, true, true, spec.allowsForageGrowthState, g_currentMission.chopperGroundLayerType)

				if realArea >  0 then

					if self.isServer then

						if growthStateArea > spec.currentGrowthStateArea then

							spec.currentGrowthStateArea = growthStateArea
							spec.currentGrowthState = growthState

						end

						spec.currentInputFruitType = fruitType

						-- ai only works on terrain detail, so we do not allow the ai to require fruits that are out of a field

						if terrainDetailPixelsSum > 0 then
							spec.currentInputFruitTypeAI = fruitType
						end

						spec.currentInputFillType = g_fruitTypeManager:getFillTypeIndexByFruitTypeIndex(fruitType)
						spec.useWindrow = false

					end

					local multiplier = g_currentMission:getHarvestScaleMultiplier(fruitType, sprayFactor, plowFactor, limeFactor, weedFactor)

					lastRealArea = realArea * multiplier
					lastThreshedArea = realArea
					lastArea = area
					spec.workAreaParameters.lastFruitType = fruitType

					break

				end

			end

		end

		if lastArea > 0 then

			if workArea.chopperAreaIndex ~= nil then

				local workArea = self:getWorkAreaByIndex(workArea.chopperAreaIndex)

				if workArea ~= nil then

					local xs,_,zs = getWorldTranslation(workArea.start)
					local xw,_,zw = getWorldTranslation(workArea.width)
					local xh,_,zh = getWorldTranslation(workArea.height)

					FSDensityMapUtil.setGroundTypeLayerArea(xs, zs, xw, zw, xh, zh, g_currentMission.chopperGroundLayerType)

				else

					workArea.chopperAreaIndex = nil
					g_logManager:xmlWarning(self.configFileName, "Invalid chopperAreaIndex '%d' for workArea '%d'!", workArea.chopperAreaIndex, workArea.index)

				end

			end

			spec.isWorking = true

		end

		spec.workAreaParameters.lastRealArea = spec.workAreaParameters.lastRealArea + lastRealArea
		spec.workAreaParameters.lastThreshedArea = spec.workAreaParameters.lastThreshedArea + lastThreshedArea
		spec.workAreaParameters.lastStatsArea = spec.workAreaParameters.lastStatsArea + lastThreshedArea
		spec.workAreaParameters.lastArea	 = spec.workAreaParameters.lastArea + lastArea

	end

	return spec.workAreaParameters.lastRealArea, spec.workAreaParameters.lastArea

end

Cutter.processCutterArea = Utils.overwrittenFunction(Cutter.processCutterArea, FruitPreparerMultiFruit.overwrite_cutter_processCutterArea )